!******************************************************************************!
!                                                                              !
!    class_particles.f90                                                       !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2020 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   class_particles.f90                            |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: J.-N. Lang, S. Rode                    |            !
!        | `---' | UFO date: 06/2020                              |            !
!        |       | UFO version: 0.9                               |            !
!        |       | Generation date: 2020-08-02 21:15             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module class_particles
  use constants_mdl
  use input_mdl, only: masscut,error_mdl,warning_mdl,nx,muMS
  implicit none
  save
  logical :: particle_model_init = .false.
  character(len=*), parameter :: modelname = "SMSP"
  character(len=*), parameter :: modelgauge = "'t Hooft-Feynman"
  character(len=*), parameter :: driver_timestamp = "Sun Aug  2 21:07:30 2020"
! number of particles in the model
  integer, parameter :: n_particles = 41
  integer, parameter :: n_orders = 3
  character(len=3), parameter, dimension(n_orders) :: order_ids                &
                  = (/'QCD', 'GC1', 'GC2'/)

! Particle:B_0, pdg_code:9000001
  integer, parameter :: P_B0 = 1
! Particle:W^+, pdg_code:24
  integer, parameter :: P_WPlus = 2
! Particle:W^-, pdg_code:-24
  integer, parameter :: P_WMinus = 3
! Particle:W_3, pdg_code:9000002
  integer, parameter :: P_W3 = 4
! Particle:g, pdg_code:21
  integer, parameter :: P_g = 5
! Particle:\nu_e, pdg_code:12
  integer, parameter :: P_nu_e = 6
! Particle:\bar{\nu}_e, pdg_code:-12
  integer, parameter :: P_nu_eAnti = 7
! Particle:\nu_\mu, pdg_code:14
  integer, parameter :: P_nu_mu = 8
! Particle:\bar{\nu}_\mu, pdg_code:-14
  integer, parameter :: P_nu_muAnti = 9
! Particle:\nu_\tau, pdg_code:16
  integer, parameter :: P_nu_tau = 10
! Particle:\bar{\nu}_\tau, pdg_code:-16
  integer, parameter :: P_nu_tauAnti = 11
! Particle:e^-, pdg_code:11
  integer, parameter :: P_eMinus = 12
! Particle:e^+, pdg_code:-11
  integer, parameter :: P_ePlus = 13
! Particle:\mu^-, pdg_code:13
  integer, parameter :: P_muMinus = 14
! Particle:\mu^+, pdg_code:-13
  integer, parameter :: P_muPlus = 15
! Particle:\tau^-, pdg_code:15
  integer, parameter :: P_tauMinus = 16
! Particle:\tau^+, pdg_code:-15
  integer, parameter :: P_tauPlus = 17
! Particle:u, pdg_code:2
  integer, parameter :: P_u = 18
! Particle:\bar{u}, pdg_code:-2
  integer, parameter :: P_uAnti = 19
! Particle:c, pdg_code:4
  integer, parameter :: P_c = 20
! Particle:\bar{c}, pdg_code:-4
  integer, parameter :: P_cAnti = 21
! Particle:t, pdg_code:6
  integer, parameter :: P_t = 22
! Particle:\bar{t}, pdg_code:-6
  integer, parameter :: P_tAnti = 23
! Particle:d, pdg_code:1
  integer, parameter :: P_d = 24
! Particle:\bar{d}, pdg_code:-1
  integer, parameter :: P_dAnti = 25
! Particle:s, pdg_code:3
  integer, parameter :: P_s = 26
! Particle:\bar{s}, pdg_code:-3
  integer, parameter :: P_sAnti = 27
! Particle:b, pdg_code:5
  integer, parameter :: P_b = 28
! Particle:\bar{b}, pdg_code:-5
  integer, parameter :: P_bAnti = 29
! Particle:\Phi_1^+, pdg_code:9000003
  integer, parameter :: P_Phi1 = 30
! Particle:\Phi_1^-, pdg_code:-9000003
  integer, parameter :: P_Phi1Anti = 31
! Particle:\Phi_2^+, pdg_code:9000004
  integer, parameter :: P_Phi2 = 32
! Particle:\Phi_2^-, pdg_code:-9000004
  integer, parameter :: P_Phi2Anti = 33
! Particle:u_{W_3}, pdg_code:9000005
  integer, parameter :: P_ghW3 = 34
! Particle:\bar{u}_{W_3}, pdg_code:-9000005
  integer, parameter :: P_ghW3Anti = 35
! Particle:u_{W^+}, pdg_code:9000006
  integer, parameter :: P_ghWp = 36
! Particle:\bar{u}_{W^+}, pdg_code:-9000006
  integer, parameter :: P_ghWpAnti = 37
! Particle:u_{W^-}, pdg_code:9000007
  integer, parameter :: P_ghWm = 38
! Particle:\bar{u}_{W^-}, pdg_code:-9000007
  integer, parameter :: P_ghWmAnti = 39
! Particle:u_{G-}, pdg_code:82
  integer, parameter :: P_ghG = 40
! Particle:\bar{u}_{G}, pdg_code:-82
  integer, parameter :: P_ghGAnti = 41
! The number of distinct mass regulators.
  integer :: n_masses

! Maximal helicity configurations. Determined by the particle with highest spin  
! e.g. massive VB  -> 3
  integer, parameter :: max_polarizations = 2

! External parameter
  complex(kind=dp) :: aS = 0.1184d0
  complex(kind=dp) :: g1 = 0.1d0
  complex(kind=dp) :: g2 = 0.02d0

! Internal parameter
  complex(kind=dp) :: ZERO
  complex(kind=dp) :: g1t
  complex(kind=dp) :: gs

! cmass2/cmass2_reg relates the mass id to the actual complex squared mass(_reg) 
! value.
  complex(kind=dp), allocatable :: cmass2(:), cmass2_reg(:)

  contains
    function sec(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = 1/cos(a)
    end function sec

    function csc(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = 1/sin(a)
    end function csc

    function cot(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = 1/tan(a)
    end function cot

    function conjugate(a) result(b)
      complex(kind=dp), intent(in) :: a
      complex(kind=dp) :: b
      b = dconjg(a)
    end function conjugate

    function get_modelname_mdl() result(ret_modelname)
      character(len=100) :: ret_modelname
      ret_modelname = modelname
    end function get_modelname_mdl

    function get_modelgauge_mdl() result(ret_modelgauge)
      character(len=100) :: ret_modelgauge
      ret_modelgauge = modelgauge
    end function get_modelgauge_mdl

    function get_driver_timestamp_mdl() result(ret_driver_timestamp)
      character(len=200) :: ret_driver_timestamp
      ret_driver_timestamp = driver_timestamp
    end function get_driver_timestamp_mdl

    function get_cmass2_mdl(m_id) result(ret_cmass2)
      integer, intent(in) :: m_id
      complex(kind=dp) :: ret_cmass2
      ret_cmass2 = cmass2(m_id)
    end function get_cmass2_mdl

    function get_particle_cmass2_mdl(p_id) result(ret_cmass2)
      integer, intent(in) :: p_id
      integer :: m_id
      complex(kind=dp) :: ret_cmass2
      m_id = get_particle_mass_id_mdl(p_id)
      ret_cmass2 = cmass2(m_id)
    end function get_particle_cmass2_mdl

    function get_cmass2_reg_mdl(m_id) result(ret_cmass2_reg)
      integer, intent(in) :: m_id
      complex(kind=dp) :: ret_cmass2_reg
      ret_cmass2_reg = cmass2_reg(m_id)
    end function get_cmass2_reg_mdl

    function get_particle_cmass2_reg_mdl(p_id) result(ret_cmass2_reg)
      integer, intent(in) :: p_id
      integer :: m_id
      complex(kind=dp) :: ret_cmass2_reg
      m_id = get_particle_mass_id_mdl(p_id)
      ret_cmass2_reg = cmass2_reg(m_id)
    end function get_particle_cmass2_reg_mdl

    function get_n_masses_mdl() result(ret_n_masses)
      integer :: ret_n_masses
      ret_n_masses = n_masses
    end function get_n_masses_mdl

    function get_n_particles_mdl() result(ret_n_particles)
      integer :: ret_n_particles
      ret_n_particles = n_particles
    end function get_n_particles_mdl

    function get_order_id_mdl(o_id) result(ret_order)
      integer, intent(in) :: o_id
      character(len=3) :: ret_order
      ret_order = order_ids(o_id)
    end function get_order_id_mdl

    function get_n_orders_mdl() result(ret_n_orders)
      integer :: ret_n_orders
      ret_n_orders = size(order_ids)
    end function get_n_orders_mdl

    function get_max_polarizations_mdl() result(ret_max_polarizations)
      integer :: ret_max_polarizations
      ret_max_polarizations = max_polarizations
    end function get_max_polarizations_mdl

    function get_particle_polarizations_mdl(p) result(npolarizations)
      integer, intent(in) :: p
      integer :: npolarizations
      character(len=3) :: p_str
      select case(get_particle_mass_reg_mdl(p))
      case(1,2)
        select case(get_particle_spin_mdl(p))
        case(-1,1)
          npolarizations = 1
        case(2,3)
          npolarizations = 2
        case default
          write(p_str,'(i3)') p
          call error_mdl("Spin other than -1,1,2,3 not supported. P_id:" // &
                        trim(p_str), where='get_particle_polarizations_mdl')
        end select
      case(3)
        select case(get_particle_spin_mdl(p))
        case(-1,1)
          npolarizations = 1
        case(2)
          npolarizations = 2
        case(3)
          npolarizations = 3
        case default
          write(p_str,'(i3)') p
          call error_mdl("Spin other than -1,1,2,3 not supported. P_id:" // &
                        trim(p_str), where='get_particle_polarizations_mdl')
          stop
        end select
      case default
        write(p_str,'(i3)') p
        call error_mdl("Mass reg cannot be other than 1,2 or 3. P_id:" // &
                      trim(p_str), where='get_particle_polarizations_mdl')
        stop
      end select
    end function get_particle_polarizations_mdl

    function is_particle_model_init_mdl() result(ret_init)
      logical :: ret_init
      ret_init = particle_model_init
    end function is_particle_model_init_mdl

    subroutine set_particle_model_init_mdl(set_init)
      logical, intent(in) :: set_init
      particle_model_init = set_init
    end subroutine set_particle_model_init_mdl

    subroutine clear_particles_mdl
      if (allocated(cmass2_reg)) deallocate(cmass2_reg)
      if (allocated(cmass2_reg)) deallocate(cmass2_reg)
      particle_model_init = .false.
    end subroutine clear_particles_mdl

    subroutine set_masscut_mdl (m)
      ! Sets mass limit below which particles are considered massless.
      real(dp), intent(in) :: m

      masscut = m
      if (is_particle_model_init_mdl()) then
        call warning_mdl('Masscut needs to be set before defining processes.',&
                        where='set_masscut_mdl')
      end if

    end subroutine set_masscut_mdl

    subroutine set_light_particle_mdl(p)
      character(*), intent(in) :: p
      call set_light_particle_value_mdl(p, .true.)
    end subroutine set_light_particle_mdl

    subroutine unset_light_particle_mdl(p)
      character(*), intent(in) :: p
      call set_light_particle_value_mdl(p, .false.)
    end subroutine unset_light_particle_mdl

    subroutine set_resonant_particle_mdl(p)
      character(*), intent(in) :: p
      call set_resonant_particle_value_mdl(p, .true.)
    end subroutine set_resonant_particle_mdl

    subroutine unset_resonant_particle_mdl(p)
      character(*), intent(in) :: p
      call set_resonant_particle_value_mdl(p, .false.)
    end subroutine unset_resonant_particle_mdl

    subroutine init_cmasses_mdl()
!     Init mass squared parameters

!     Start with mass Ids and complex masses
!     The mass ids are given by their order in UFO
      n_masses = 2

      if (allocated(cmass2)) deallocate(cmass2)
      allocate(cmass2(n_masses))
      if (allocated(cmass2_reg)) deallocate(cmass2_reg)
      allocate(cmass2_reg(n_masses))
      cmass2(1:2) = cnul
      cmass2_reg(1:2) = cnul



      call init_parameters_mdl()

    end subroutine init_cmasses_mdl

    subroutine init_parameters_mdl()
      ZERO = 0d0
      g1t = -g1/2d0
      gs = 2d0*sqrt(pi)*sqrt(aS)
    end subroutine init_parameters_mdl

    subroutine init_particles_mdl()
      call init_cmasses_mdl
    end subroutine init_particles_mdl

    function get_doublet_partner_mdl(particle_in) result(particle_out)
      integer :: particle_in, particle_out
      select case(particle_in)
      case(P_nu_e)
        particle_out = P_eMinus
      case(P_nu_mu)
        particle_out = P_muMinus
      case(P_nu_tau)
        particle_out = P_tauMinus
      case(P_u)
        particle_out = P_d
      case(P_c)
        particle_out = P_s
      case(P_t)
        particle_out = P_b
      case(P_eMinus)
        particle_out = P_nu_e
      case(P_nu_eAnti)
        particle_out = P_ePlus
      case(P_ePlus)
        particle_out = P_nu_eAnti
      case(P_muMinus)
        particle_out = P_nu_mu
      case(P_nu_muAnti)
        particle_out = P_muPlus
      case(P_muPlus)
        particle_out = P_nu_muAnti
      case(P_tauMinus)
        particle_out = P_nu_tau
      case(P_nu_tauAnti)
        particle_out = P_tauPlus
      case(P_tauPlus)
        particle_out = P_nu_tauAnti
      case(P_d)
        particle_out = P_u
      case(P_uAnti)
        particle_out = P_dAnti
      case(P_dAnti)
        particle_out = P_uAnti
      case(P_s)
        particle_out = P_c
      case(P_cAnti)
        particle_out = P_sAnti
      case(P_sAnti)
        particle_out = P_cAnti
      case(P_b)
        particle_out = P_t
      case(P_tAnti)
        particle_out = P_bAnti
      case(P_bAnti)
        particle_out = P_tAnti
      case default
        particle_out = 0
      end select 
    end function get_doublet_partner_mdl

    function is_firstgen_mdl(particle) result(isFirstGen)
      integer :: particle
      logical :: isFirstGen
      select case(particle)
      case(P_eMinus, P_ePlus, P_nu_e, P_nu_eAnti, P_d, P_dAnti, P_u, P_uAnti)
        isFirstGen = .true.
      case default
        isFirstGen = .false.
      end select
    end function is_firstgen_mdl

    function get_secondgen_mdl(particle) result(secondGen)
      integer :: particle, secondGen
      select case(particle)
      case(P_eMinus)
        secondGen = P_muMinus
      case(P_ePlus)
        secondGen = P_muPlus
      case(P_nu_e)
        secondGen = P_nu_mu
      case(P_nu_eAnti)
        secondGen = P_nu_muAnti
      case(P_d)
        secondGen = P_s
      case(P_dAnti)
        secondGen = P_sAnti
      case(P_u)
        secondGen = P_c
      case(P_uAnti)
        secondGen = P_cAnti
      end select
    end function get_secondgen_mdl

    function get_thirdgen_mdl(particle) result(thirdGen)
      integer :: particle, thirdGen
      select case(particle)
      case(P_eMinus)
        thirdGen = P_tauMinus
      case(P_ePlus)
        thirdGen = P_tauPlus
      case(P_nu_e)
        thirdGen = P_nu_tau
      case(P_nu_eAnti)
        thirdGen = P_nu_tauAnti
      case(P_d)
        thirdGen = P_b
      case(P_dAnti)
        thirdGen = P_bAnti
      case(P_u)
        thirdGen = P_t
      case(P_uAnti)
        thirdGen = P_tAnti
      end select
    end function get_thirdgen_mdl

    function get_particle_id_mdl(p_str) result(particle_id)
      character(*), intent(in) :: p_str
      integer :: particle_id
      select case(p_str)
      case("B0")
        particle_id = 1
      case("W+")
        particle_id = 2
      case("W-")
        particle_id = 3
      case("W3")
        particle_id = 4
      case("g")
        particle_id = 5
      case("nu_e")
        particle_id = 6
      case("nu_e~")
        particle_id = 7
      case("nu_mu")
        particle_id = 8
      case("nu_mu~")
        particle_id = 9
      case("nu_tau")
        particle_id = 10
      case("nu_tau~")
        particle_id = 11
      case("e-")
        particle_id = 12
      case("e+")
        particle_id = 13
      case("mu-")
        particle_id = 14
      case("mu+")
        particle_id = 15
      case("tau-")
        particle_id = 16
      case("tau+")
        particle_id = 17
      case("u")
        particle_id = 18
      case("u~")
        particle_id = 19
      case("c")
        particle_id = 20
      case("c~")
        particle_id = 21
      case("t")
        particle_id = 22
      case("t~")
        particle_id = 23
      case("d")
        particle_id = 24
      case("d~")
        particle_id = 25
      case("s")
        particle_id = 26
      case("s~")
        particle_id = 27
      case("b")
        particle_id = 28
      case("b~")
        particle_id = 29
      case("Phi1")
        particle_id = 30
      case("Phi1~")
        particle_id = 31
      case("Phi2")
        particle_id = 32
      case("Phi2~")
        particle_id = 33
      case("ghW3")
        particle_id = 34
      case("ghW3~")
        particle_id = 35
      case("ghWp")
        particle_id = 36
      case("ghWp~")
        particle_id = 37
      case("ghWm")
        particle_id = 38
      case("ghWm~")
        particle_id = 39
      case("ghG")
        particle_id = 40
      case("ghG~")
        particle_id = 41
      end select 
    end function get_particle_id_mdl

    function get_particle_name_mdl(p) result(pname)
      integer, intent(in) :: p
      character(len=10) :: pname
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        pname = "B0"
      case(P_WPlus)
        pname = "W+"
      case(P_WMinus)
        pname = "W-"
      case(P_W3)
        pname = "W3"
      case(P_g)
        pname = "g"
      case(P_nu_e)
        pname = "nu_e"
      case(P_nu_eAnti)
        pname = "nu_e~"
      case(P_nu_mu)
        pname = "nu_mu"
      case(P_nu_muAnti)
        pname = "nu_mu~"
      case(P_nu_tau)
        pname = "nu_tau"
      case(P_nu_tauAnti)
        pname = "nu_tau~"
      case(P_eMinus)
        pname = "e-"
      case(P_ePlus)
        pname = "e+"
      case(P_muMinus)
        pname = "mu-"
      case(P_muPlus)
        pname = "mu+"
      case(P_tauMinus)
        pname = "tau-"
      case(P_tauPlus)
        pname = "tau+"
      case(P_u)
        pname = "u"
      case(P_uAnti)
        pname = "u~"
      case(P_c)
        pname = "c"
      case(P_cAnti)
        pname = "c~"
      case(P_t)
        pname = "t"
      case(P_tAnti)
        pname = "t~"
      case(P_d)
        pname = "d"
      case(P_dAnti)
        pname = "d~"
      case(P_s)
        pname = "s"
      case(P_sAnti)
        pname = "s~"
      case(P_b)
        pname = "b"
      case(P_bAnti)
        pname = "b~"
      case(P_Phi1)
        pname = "Phi1"
      case(P_Phi1Anti)
        pname = "Phi1~"
      case(P_Phi2)
        pname = "Phi2"
      case(P_Phi2Anti)
        pname = "Phi2~"
      case(P_ghW3)
        pname = "ghW3"
      case(P_ghW3Anti)
        pname = "ghW3~"
      case(P_ghWp)
        pname = "ghWp"
      case(P_ghWpAnti)
        pname = "ghWp~"
      case(P_ghWm)
        pname = "ghWm"
      case(P_ghWmAnti)
        pname = "ghWm~"
      case(P_ghG)
        pname = "ghG"
      case(P_ghGAnti)
        pname = "ghG~"
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_name_mdl")
      end select 
    end function get_particle_name_mdl

    function get_particle_texname_mdl(p) result(ptexname)
      integer, intent(in) :: p
      character(len=30) :: ptexname
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        ptexname = "B_0"
      case(P_WPlus)
        ptexname = "W^+"
      case(P_WMinus)
        ptexname = "W^-"
      case(P_W3)
        ptexname = "W_3"
      case(P_g)
        ptexname = "g"
      case(P_nu_e)
        ptexname = "\nu_e"
      case(P_nu_eAnti)
        ptexname = "\bar{\nu}_e"
      case(P_nu_mu)
        ptexname = "\nu_\mu"
      case(P_nu_muAnti)
        ptexname = "\bar{\nu}_\mu"
      case(P_nu_tau)
        ptexname = "\nu_\tau"
      case(P_nu_tauAnti)
        ptexname = "\bar{\nu}_\tau"
      case(P_eMinus)
        ptexname = "e^-"
      case(P_ePlus)
        ptexname = "e^+"
      case(P_muMinus)
        ptexname = "\mu^-"
      case(P_muPlus)
        ptexname = "\mu^+"
      case(P_tauMinus)
        ptexname = "\tau^-"
      case(P_tauPlus)
        ptexname = "\tau^+"
      case(P_u)
        ptexname = "u"
      case(P_uAnti)
        ptexname = "\bar{u}"
      case(P_c)
        ptexname = "c"
      case(P_cAnti)
        ptexname = "\bar{c}"
      case(P_t)
        ptexname = "t"
      case(P_tAnti)
        ptexname = "\bar{t}"
      case(P_d)
        ptexname = "d"
      case(P_dAnti)
        ptexname = "\bar{d}"
      case(P_s)
        ptexname = "s"
      case(P_sAnti)
        ptexname = "\bar{s}"
      case(P_b)
        ptexname = "b"
      case(P_bAnti)
        ptexname = "\bar{b}"
      case(P_Phi1)
        ptexname = "\Phi_1^+"
      case(P_Phi1Anti)
        ptexname = "\Phi_1^-"
      case(P_Phi2)
        ptexname = "\Phi_2^+"
      case(P_Phi2Anti)
        ptexname = "\Phi_2^-"
      case(P_ghW3)
        ptexname = "u_{W_3}"
      case(P_ghW3Anti)
        ptexname = "\bar{u}_{W_3}"
      case(P_ghWp)
        ptexname = "u_{W^+}"
      case(P_ghWpAnti)
        ptexname = "\bar{u}_{W^+}"
      case(P_ghWm)
        ptexname = "u_{W^-}"
      case(P_ghWmAnti)
        ptexname = "\bar{u}_{W^-}"
      case(P_ghG)
        ptexname = "u_{G-}"
      case(P_ghGAnti)
        ptexname = "\bar{u}_{G}"
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_texname_mdl")
      end select 
    end function get_particle_texname_mdl

    function get_antiparticle_id_mdl(p_str) result(particle_id)
      character(*), intent(in) :: p_str
      integer :: particle_id
      select case(p_str)
      case("B0")
        particle_id = 1
      case("W-")
        particle_id = 2
      case("W+")
        particle_id = 3
      case("W3")
        particle_id = 4
      case("g")
        particle_id = 5
      case("nu_e~")
        particle_id = 6
      case("nu_e")
        particle_id = 7
      case("nu_mu~")
        particle_id = 8
      case("nu_mu")
        particle_id = 9
      case("nu_tau~")
        particle_id = 10
      case("nu_tau")
        particle_id = 11
      case("e+")
        particle_id = 12
      case("e-")
        particle_id = 13
      case("mu+")
        particle_id = 14
      case("mu-")
        particle_id = 15
      case("tau+")
        particle_id = 16
      case("tau-")
        particle_id = 17
      case("u~")
        particle_id = 18
      case("u")
        particle_id = 19
      case("c~")
        particle_id = 20
      case("c")
        particle_id = 21
      case("t~")
        particle_id = 22
      case("t")
        particle_id = 23
      case("d~")
        particle_id = 24
      case("d")
        particle_id = 25
      case("s~")
        particle_id = 26
      case("s")
        particle_id = 27
      case("b~")
        particle_id = 28
      case("b")
        particle_id = 29
      case("Phi1~")
        particle_id = 30
      case("Phi1")
        particle_id = 31
      case("Phi2~")
        particle_id = 32
      case("Phi2")
        particle_id = 33
      case("ghW3~")
        particle_id = 34
      case("ghW3")
        particle_id = 35
      case("ghWp~")
        particle_id = 36
      case("ghWp")
        particle_id = 37
      case("ghWm~")
        particle_id = 38
      case("ghWm")
        particle_id = 39
      case("ghG~")
        particle_id = 40
      case("ghG")
        particle_id = 41
      case default
        call error_mdl("Particle  "// trim(p_str) // " not defined.", where="  &
        get_antiparticle_id_mdl")
      end select 
    end function get_antiparticle_id_mdl

    function get_particle_antiparticle_mdl(p) result(ppa)
      integer, intent(in) :: p
      integer :: ppa
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        ppa = P_B0
      case(P_WPlus)
        ppa = P_WMinus
      case(P_WMinus)
        ppa = P_WPlus
      case(P_W3)
        ppa = P_W3
      case(P_g)
        ppa = P_g
      case(P_nu_e)
        ppa = P_nu_eAnti
      case(P_nu_eAnti)
        ppa = P_nu_e
      case(P_nu_mu)
        ppa = P_nu_muAnti
      case(P_nu_muAnti)
        ppa = P_nu_mu
      case(P_nu_tau)
        ppa = P_nu_tauAnti
      case(P_nu_tauAnti)
        ppa = P_nu_tau
      case(P_eMinus)
        ppa = P_ePlus
      case(P_ePlus)
        ppa = P_eMinus
      case(P_muMinus)
        ppa = P_muPlus
      case(P_muPlus)
        ppa = P_muMinus
      case(P_tauMinus)
        ppa = P_tauPlus
      case(P_tauPlus)
        ppa = P_tauMinus
      case(P_u)
        ppa = P_uAnti
      case(P_uAnti)
        ppa = P_u
      case(P_c)
        ppa = P_cAnti
      case(P_cAnti)
        ppa = P_c
      case(P_t)
        ppa = P_tAnti
      case(P_tAnti)
        ppa = P_t
      case(P_d)
        ppa = P_dAnti
      case(P_dAnti)
        ppa = P_d
      case(P_s)
        ppa = P_sAnti
      case(P_sAnti)
        ppa = P_s
      case(P_b)
        ppa = P_bAnti
      case(P_bAnti)
        ppa = P_b
      case(P_Phi1)
        ppa = P_Phi1Anti
      case(P_Phi1Anti)
        ppa = P_Phi1
      case(P_Phi2)
        ppa = P_Phi2Anti
      case(P_Phi2Anti)
        ppa = P_Phi2
      case(P_ghW3)
        ppa = P_ghW3Anti
      case(P_ghW3Anti)
        ppa = P_ghW3
      case(P_ghWp)
        ppa = P_ghWpAnti
      case(P_ghWpAnti)
        ppa = P_ghWp
      case(P_ghWm)
        ppa = P_ghWmAnti
      case(P_ghWmAnti)
        ppa = P_ghWm
      case(P_ghG)
        ppa = P_ghGAnti
      case(P_ghGAnti)
        ppa = P_ghG
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_antiparticle_mdl")
      end select 
    end function get_particle_antiparticle_mdl

    function get_particle_spin_mdl(p) result(pspin)
      integer, intent(in) :: p
      integer :: pspin
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        pspin = 3
      case(P_WPlus)
        pspin = 3
      case(P_WMinus)
        pspin = 3
      case(P_W3)
        pspin = 3
      case(P_g)
        pspin = 3
      case(P_nu_e)
        pspin = 2
      case(P_nu_eAnti)
        pspin = 2
      case(P_nu_mu)
        pspin = 2
      case(P_nu_muAnti)
        pspin = 2
      case(P_nu_tau)
        pspin = 2
      case(P_nu_tauAnti)
        pspin = 2
      case(P_eMinus)
        pspin = 2
      case(P_ePlus)
        pspin = 2
      case(P_muMinus)
        pspin = 2
      case(P_muPlus)
        pspin = 2
      case(P_tauMinus)
        pspin = 2
      case(P_tauPlus)
        pspin = 2
      case(P_u)
        pspin = 2
      case(P_uAnti)
        pspin = 2
      case(P_c)
        pspin = 2
      case(P_cAnti)
        pspin = 2
      case(P_t)
        pspin = 2
      case(P_tAnti)
        pspin = 2
      case(P_d)
        pspin = 2
      case(P_dAnti)
        pspin = 2
      case(P_s)
        pspin = 2
      case(P_sAnti)
        pspin = 2
      case(P_b)
        pspin = 2
      case(P_bAnti)
        pspin = 2
      case(P_Phi1)
        pspin = 1
      case(P_Phi1Anti)
        pspin = 1
      case(P_Phi2)
        pspin = 1
      case(P_Phi2Anti)
        pspin = 1
      case(P_ghW3)
        pspin = -1
      case(P_ghW3Anti)
        pspin = -1
      case(P_ghWp)
        pspin = -1
      case(P_ghWpAnti)
        pspin = -1
      case(P_ghWm)
        pspin = -1
      case(P_ghWmAnti)
        pspin = -1
      case(P_ghG)
        pspin = -1
      case(P_ghGAnti)
        pspin = -1
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_spin_mdl")
      end select 
    end function get_particle_spin_mdl

    function get_particle_colour_mdl(p) result(pcolour)
      integer, intent(in) :: p
      integer :: pcolour
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        pcolour = 1
      case(P_WPlus)
        pcolour = 1
      case(P_WMinus)
        pcolour = 1
      case(P_W3)
        pcolour = 1
      case(P_g)
        pcolour = 8
      case(P_nu_e)
        pcolour = 1
      case(P_nu_eAnti)
        pcolour = 1
      case(P_nu_mu)
        pcolour = 1
      case(P_nu_muAnti)
        pcolour = 1
      case(P_nu_tau)
        pcolour = 1
      case(P_nu_tauAnti)
        pcolour = 1
      case(P_eMinus)
        pcolour = 1
      case(P_ePlus)
        pcolour = 1
      case(P_muMinus)
        pcolour = 1
      case(P_muPlus)
        pcolour = 1
      case(P_tauMinus)
        pcolour = 1
      case(P_tauPlus)
        pcolour = 1
      case(P_u)
        pcolour = 3
      case(P_uAnti)
        pcolour = -3
      case(P_c)
        pcolour = 3
      case(P_cAnti)
        pcolour = -3
      case(P_t)
        pcolour = 3
      case(P_tAnti)
        pcolour = -3
      case(P_d)
        pcolour = 3
      case(P_dAnti)
        pcolour = -3
      case(P_s)
        pcolour = 3
      case(P_sAnti)
        pcolour = -3
      case(P_b)
        pcolour = 3
      case(P_bAnti)
        pcolour = -3
      case(P_Phi1)
        pcolour = 1
      case(P_Phi1Anti)
        pcolour = 1
      case(P_Phi2)
        pcolour = 1
      case(P_Phi2Anti)
        pcolour = 1
      case(P_ghW3)
        pcolour = 1
      case(P_ghW3Anti)
        pcolour = 1
      case(P_ghWp)
        pcolour = 1
      case(P_ghWpAnti)
        pcolour = 1
      case(P_ghWm)
        pcolour = 1
      case(P_ghWmAnti)
        pcolour = 1
      case(P_ghG)
        pcolour = 8
      case(P_ghGAnti)
        pcolour = 8
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="get_particle_colour_mdl")
      end select 
    end function get_particle_colour_mdl

    function get_mass_name_mdl(mass_id) result(mass_name)
      integer, intent(in) :: mass_id
      character(len=6) :: mass_name, mass_id_str
      if (mass_id .le. 2) then
        mass_name = "M0"
      else
        write(mass_id_str,'(i3)') mass_id
        call error_mdl("Mass id " // trim(mass_id_str) // " not defined.",     &
        where="get_mass_name_mdl")
      end if 
    end function get_mass_name_mdl

    function get_particle_mass_id_mdl(p) result(pmassid)
      integer, intent(in) :: p
      integer :: pmassid
      select case(p)
      case(P_B0)
        pmassid = 1
      case(P_WPlus)
        pmassid = 1
      case(P_WMinus)
        pmassid = 1
      case(P_W3)
        pmassid = 1
      case(P_g)
        pmassid = 1
      case(P_nu_e)
        pmassid = 1
      case(P_nu_eAnti)
        pmassid = 1
      case(P_nu_mu)
        pmassid = 1
      case(P_nu_muAnti)
        pmassid = 1
      case(P_nu_tau)
        pmassid = 1
      case(P_nu_tauAnti)
        pmassid = 1
      case(P_eMinus)
        pmassid = 1
      case(P_ePlus)
        pmassid = 1
      case(P_muMinus)
        pmassid = 1
      case(P_muPlus)
        pmassid = 1
      case(P_tauMinus)
        pmassid = 1
      case(P_tauPlus)
        pmassid = 1
      case(P_u)
        pmassid = 1
      case(P_uAnti)
        pmassid = 1
      case(P_c)
        pmassid = 1
      case(P_cAnti)
        pmassid = 1
      case(P_t)
        pmassid = 1
      case(P_tAnti)
        pmassid = 1
      case(P_d)
        pmassid = 1
      case(P_dAnti)
        pmassid = 1
      case(P_s)
        pmassid = 1
      case(P_sAnti)
        pmassid = 1
      case(P_b)
        pmassid = 1
      case(P_bAnti)
        pmassid = 1
      case(P_Phi1)
        pmassid = 1
      case(P_Phi1Anti)
        pmassid = 1
      case(P_Phi2)
        pmassid = 1
      case(P_Phi2Anti)
        pmassid = 1
      case(P_ghW3)
        pmassid = 1
      case(P_ghW3Anti)
        pmassid = 1
      case(P_ghWp)
        pmassid = 1
      case(P_ghWpAnti)
        pmassid = 1
      case(P_ghWm)
        pmassid = 1
      case(P_ghWmAnti)
        pmassid = 1
      case(P_ghG)
        pmassid = 1
      case(P_ghGAnti)
        pmassid = 1
      end select 
    end function get_particle_mass_id_mdl

    function get_particle_mass_reg_mdl(p) result(pmassreg)
      integer, intent(in) :: p
      integer :: pmassreg
      select case(p)
      case(P_B0)
        pmassreg = 1
      case(P_WPlus)
        pmassreg = 1
      case(P_WMinus)
        pmassreg = 1
      case(P_W3)
        pmassreg = 1
      case(P_g)
        pmassreg = 1
      case(P_nu_e)
        pmassreg = 1
      case(P_nu_eAnti)
        pmassreg = 1
      case(P_nu_mu)
        pmassreg = 1
      case(P_nu_muAnti)
        pmassreg = 1
      case(P_nu_tau)
        pmassreg = 1
      case(P_nu_tauAnti)
        pmassreg = 1
      case(P_eMinus)
        pmassreg = 1
      case(P_ePlus)
        pmassreg = 1
      case(P_muMinus)
        pmassreg = 1
      case(P_muPlus)
        pmassreg = 1
      case(P_tauMinus)
        pmassreg = 1
      case(P_tauPlus)
        pmassreg = 1
      case(P_u)
        pmassreg = 1
      case(P_uAnti)
        pmassreg = 1
      case(P_c)
        pmassreg = 1
      case(P_cAnti)
        pmassreg = 1
      case(P_t)
        pmassreg = 1
      case(P_tAnti)
        pmassreg = 1
      case(P_d)
        pmassreg = 1
      case(P_dAnti)
        pmassreg = 1
      case(P_s)
        pmassreg = 1
      case(P_sAnti)
        pmassreg = 1
      case(P_b)
        pmassreg = 1
      case(P_bAnti)
        pmassreg = 1
      case(P_Phi1)
        pmassreg = 1
      case(P_Phi1Anti)
        pmassreg = 1
      case(P_Phi2)
        pmassreg = 1
      case(P_Phi2Anti)
        pmassreg = 1
      case(P_ghW3)
        pmassreg = 1
      case(P_ghW3Anti)
        pmassreg = 1
      case(P_ghWp)
        pmassreg = 1
      case(P_ghWpAnti)
        pmassreg = 1
      case(P_ghWm)
        pmassreg = 1
      case(P_ghWmAnti)
        pmassreg = 1
      case(P_ghG)
        pmassreg = 1
      case(P_ghGAnti)
        pmassreg = 1
      end select 
    end function get_particle_mass_reg_mdl

    function get_particle_width_mdl(p) result(pwidth)
      integer, intent(in) :: p
      complex(kind=dp) :: pwidth
      select case(p)
      case default
        pwidth = cnul
      end select 
    end function get_particle_width_mdl

    subroutine set_parameter_mdl(param,value)
      character(*), intent(in) :: param
      complex(dp), intent(in) :: value
      if (particle_model_init) then
        call warning_mdl                                                       &
        ('Parameters need to be set before the process  generation.', where    &
        ='set_parameter_mdl')
      end if 
      select case(param)
        case("muMS")
          muMS = real(value,kind=dp)
        case("g2")
          g2 = real(value,kind=dp)
        case("g1")
          g1 = real(value,kind=dp)
        case("aS")
          aS = real(value,kind=dp)
        case default
        call warning_mdl('Trying to change unknown/internal parameter:' // trim&
        (param), where='set_parameter_mdl')
      end select 
    end subroutine set_parameter_mdl

    function get_parameter_mdl(param)
      character(*), intent(in) :: param
      complex(dp) :: get_parameter_mdl
      select case(param)
        case("muMS")
          get_parameter_mdl = muMS
        case("g2")
          get_parameter_mdl = g2
        case("g1")
          get_parameter_mdl = g1
        case("aS")
          get_parameter_mdl = aS
        case("ZERO")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = ZERO
        case("gs")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = gs
        case("g1t")
          call warning_mdl("Trying to get an internal variable which might not &
          be computed yet", where="get_parameter_mdl")
          get_parameter_mdl = g1t
        case default
          call warning_mdl('Trying to get an unknown parameter: ' // trim(param&
          ), where='get_parameter_mdl')
          get_parameter_mdl = cnul
      end select 
    end function get_parameter_mdl

    subroutine print_parameters_mdl(level)
      integer, intent(in) :: level
      character(len=15) :: fmts = '(4x,a10,g21.14)'
      character(len=24) :: fmtsc = '(4x,a10,g21.14,x,g21.14)'
      character(len=15) :: fmtm = '(4x,a7,es23.16)'
      character(len=25) :: fmtc = '(4x,a7,es23.16,x,es23.16)'
      character(len=26) :: fmtmw = '(4x,a7,es23.16,a8,es23.16)'
      character(len=18) :: fmtml = '(4x,a7,es23.16,a8)'
      write(nx,'(1x,75("-"))')
      write(nx,*) 'Input: Scale values:'
      write(nx,fmts) adjustl('muMS ='),real(muMS)
      write(nx,'(1x,75("-"))')
      write(nx,*) 'Input: Coupling values:'
      write(nx,fmtm) adjustl('aS ='),real(aS)
      write(nx,fmtm) adjustl('g1 ='),real(g1)
      write(nx,fmtm) adjustl('g2 ='),real(g2)
      write(nx,'(1x,75("-"))')
      write(nx,*) 'Input: Pole masses and widths [GeV]:'
      write(nx,'(1x,75("-"))')
      if (level .gt. 0) then
        write(nx,*) 'Complex masses squared [GeV^2] (internal):'
        write(nx,'(1x,75("-"))')
        write(nx,*) 'Internal parameter values:'
        write(nx,fmtc) 'ZERO =',ZERO
        write(nx,fmtc) 'g1t =',g1t
        write(nx,fmtc) 'gs =',gs
        write(nx,'(1x,75("-"))')
      end if 
    end subroutine print_parameters_mdl

    subroutine set_resonant_particle_value_mdl(p,val)
      character(*), intent(in) :: p
      logical, intent(in) :: val
      if (particle_model_init) then
        call warning_mdl                                                       &
        ('Resonances need to be set before defining processes.', where         &
        ='set_resonant_particle_value_mdl')
      end if 
      select case(p)
      case default
        call warning_mdl                                                       &
        ('Trying to set resonance for non-supported particle: ' // trim(p),    &
        where='set_resonant_particle_value_mdl')
      end select 
    end subroutine set_resonant_particle_value_mdl

    function is_resonant_particle_id_mdl(p) result(is_resonant_particle)
      integer, intent(in) :: p
      logical :: is_resonant_particle
      select case(p)
      case default
        call warning_mdl                                                       &
        ('Trying to determine resonance status for non-support particle id.',  &
        where='is_resonant_particle_id_mdl')
        is_resonant_particle = .false.
      end select 
    end function is_resonant_particle_id_mdl

    subroutine set_light_particle_value_mdl(p,val)
      use input_mdl, only: warning_mdl
      character(*), intent(in) :: p
      logical, intent(in) :: val
      if (particle_model_init) then
        call warning_mdl                                                       &
        ('Light particles need to be set before defining processes.', where    &
        ='set_light_particle_value_mdl')
      end if 
      select case(p)
      case default
        call warning_mdl                                                       &
        ('Trying to set light particle for non-supported particle: ' // trim(p &
        ), where='set_light_particle_value_mdl')
      end select 
    end subroutine set_light_particle_value_mdl

    function is_light_mass_mdl(m)
      use input_mdl, only: warning_mdl
      character(*), intent(in) :: m
      logical :: is_light_mass_mdl
      select case(m)
      case default
        call warning_mdl                                                       &
        ('Trying to determine light status for non-support mass: ' // trim(m), &
        where='is_light_mass_mdl')
        is_light_mass_mdl = .false.
      end select 
    end function is_light_mass_mdl

    function is_particle_mdl(p) result(is_particle)
      integer, intent(in) :: p
      logical :: is_particle
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        is_particle = .true.
      case(P_WPlus)
        is_particle = .false.
      case(P_WMinus)
        is_particle = .true.
      case(P_W3)
        is_particle = .true.
      case(P_g)
        is_particle = .true.
      case(P_nu_e)
        is_particle = .true.
      case(P_nu_eAnti)
        is_particle = .false.
      case(P_nu_mu)
        is_particle = .true.
      case(P_nu_muAnti)
        is_particle = .false.
      case(P_nu_tau)
        is_particle = .true.
      case(P_nu_tauAnti)
        is_particle = .false.
      case(P_eMinus)
        is_particle = .true.
      case(P_ePlus)
        is_particle = .false.
      case(P_muMinus)
        is_particle = .true.
      case(P_muPlus)
        is_particle = .false.
      case(P_tauMinus)
        is_particle = .true.
      case(P_tauPlus)
        is_particle = .false.
      case(P_u)
        is_particle = .true.
      case(P_uAnti)
        is_particle = .false.
      case(P_c)
        is_particle = .true.
      case(P_cAnti)
        is_particle = .false.
      case(P_t)
        is_particle = .true.
      case(P_tAnti)
        is_particle = .false.
      case(P_d)
        is_particle = .true.
      case(P_dAnti)
        is_particle = .false.
      case(P_s)
        is_particle = .true.
      case(P_sAnti)
        is_particle = .false.
      case(P_b)
        is_particle = .true.
      case(P_bAnti)
        is_particle = .false.
      case(P_Phi1)
        is_particle = .true.
      case(P_Phi1Anti)
        is_particle = .false.
      case(P_Phi2)
        is_particle = .true.
      case(P_Phi2Anti)
        is_particle = .false.
      case(P_ghW3)
        is_particle = .true.
      case(P_ghW3Anti)
        is_particle = .false.
      case(P_ghWp)
        is_particle = .true.
      case(P_ghWpAnti)
        is_particle = .false.
      case(P_ghWm)
        is_particle = .true.
      case(P_ghWmAnti)
        is_particle = .false.
      case(P_ghG)
        is_particle = .true.
      case(P_ghGAnti)
        is_particle = .false.
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="is_particle_mdl")
      end select 
    end function is_particle_mdl

    function is_backgroundfield_mdl(p) result(isbf)
      integer, intent(in) :: p
      logical :: isbf
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        isbf = .true.
      case(P_WPlus)
        isbf = .true.
      case(P_WMinus)
        isbf = .true.
      case(P_W3)
        isbf = .true.
      case(P_g)
        isbf = .true.
      case(P_nu_e)
        isbf = .true.
      case(P_nu_eAnti)
        isbf = .true.
      case(P_nu_mu)
        isbf = .true.
      case(P_nu_muAnti)
        isbf = .true.
      case(P_nu_tau)
        isbf = .true.
      case(P_nu_tauAnti)
        isbf = .true.
      case(P_eMinus)
        isbf = .true.
      case(P_ePlus)
        isbf = .true.
      case(P_muMinus)
        isbf = .true.
      case(P_muPlus)
        isbf = .true.
      case(P_tauMinus)
        isbf = .true.
      case(P_tauPlus)
        isbf = .true.
      case(P_u)
        isbf = .true.
      case(P_uAnti)
        isbf = .true.
      case(P_c)
        isbf = .true.
      case(P_cAnti)
        isbf = .true.
      case(P_t)
        isbf = .true.
      case(P_tAnti)
        isbf = .true.
      case(P_d)
        isbf = .true.
      case(P_dAnti)
        isbf = .true.
      case(P_s)
        isbf = .true.
      case(P_sAnti)
        isbf = .true.
      case(P_b)
        isbf = .true.
      case(P_bAnti)
        isbf = .true.
      case(P_Phi1)
        isbf = .true.
      case(P_Phi1Anti)
        isbf = .true.
      case(P_Phi2)
        isbf = .true.
      case(P_Phi2Anti)
        isbf = .true.
      case(P_ghW3)
        isbf = .true.
      case(P_ghW3Anti)
        isbf = .true.
      case(P_ghWp)
        isbf = .true.
      case(P_ghWpAnti)
        isbf = .true.
      case(P_ghWm)
        isbf = .true.
      case(P_ghWmAnti)
        isbf = .true.
      case(P_ghG)
        isbf = .true.
      case(P_ghGAnti)
        isbf = .true.
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="is_backgroundfield_mdl")
      end select 
    end function is_backgroundfield_mdl

    function is_quantumfield_mdl(p) result(isqf)
      integer, intent(in) :: p
      logical :: isqf
      character(len=3) :: p_str
      select case(p)
      case(P_B0)
        isqf = .true.
      case(P_WPlus)
        isqf = .true.
      case(P_WMinus)
        isqf = .true.
      case(P_W3)
        isqf = .true.
      case(P_g)
        isqf = .true.
      case(P_nu_e)
        isqf = .true.
      case(P_nu_eAnti)
        isqf = .true.
      case(P_nu_mu)
        isqf = .true.
      case(P_nu_muAnti)
        isqf = .true.
      case(P_nu_tau)
        isqf = .true.
      case(P_nu_tauAnti)
        isqf = .true.
      case(P_eMinus)
        isqf = .true.
      case(P_ePlus)
        isqf = .true.
      case(P_muMinus)
        isqf = .true.
      case(P_muPlus)
        isqf = .true.
      case(P_tauMinus)
        isqf = .true.
      case(P_tauPlus)
        isqf = .true.
      case(P_u)
        isqf = .true.
      case(P_uAnti)
        isqf = .true.
      case(P_c)
        isqf = .true.
      case(P_cAnti)
        isqf = .true.
      case(P_t)
        isqf = .true.
      case(P_tAnti)
        isqf = .true.
      case(P_d)
        isqf = .true.
      case(P_dAnti)
        isqf = .true.
      case(P_s)
        isqf = .true.
      case(P_sAnti)
        isqf = .true.
      case(P_b)
        isqf = .true.
      case(P_bAnti)
        isqf = .true.
      case(P_Phi1)
        isqf = .true.
      case(P_Phi1Anti)
        isqf = .true.
      case(P_Phi2)
        isqf = .true.
      case(P_Phi2Anti)
        isqf = .true.
      case(P_ghW3)
        isqf = .true.
      case(P_ghW3Anti)
        isqf = .true.
      case(P_ghWp)
        isqf = .true.
      case(P_ghWpAnti)
        isqf = .true.
      case(P_ghWm)
        isqf = .true.
      case(P_ghWmAnti)
        isqf = .true.
      case(P_ghG)
        isqf = .true.
      case(P_ghGAnti)
        isqf = .true.
      case default
        write(p_str,'(i3)') p
        call error_mdl("Particle id " // trim(p_str) // " not defined.", where &
        ="is_quantumfield_mdl")
      end select 
    end function is_quantumfield_mdl

    function has_feature_mdl(feature) result(has_feature)
      character(*), intent(in) :: feature
      logical :: has_feature
      select case(feature)
      case("fermionloop_opt")
        has_feature = .true.
      case("qcd_rescaling")
        has_feature = .true.
      case default
        has_feature = .false.
      end select 
    end function has_feature_mdl

    function get_particle_type_mdl(particle_in) result(particle_type)
      integer, intent(in) :: particle_in
      character(2) :: particle_type
      select case (get_particle_spin_mdl(particle_in))
      ! ghost
      case(-1)
        select case (is_particle_mdl(particle_in))
        ! ghost
        case (.true.)
          particle_type = 'x'
        ! antighost
        case (.false.)
          particle_type = 'x~'
        end select
      ! scalar
      case (1)
        particle_type = 's'
      ! fermion
      case (2)
        select case (is_particle_mdl(particle_in))
        ! fermion
        case (.true.)
          particle_type = 'f'
        ! antifermion
        case (.false.)
          particle_type = 'f~'
        end select
      ! vector
      case (3)
        particle_type = 'v'
      case default
        particle_type = '0'
      end select
    end function get_particle_type_mdl

    function get_particle_type2_mdl(particle_in) result(particle_type)
      integer, intent(in) :: particle_in
      character(2) :: particle_type
      select case (get_particle_spin_mdl(particle_in))
      ! ghost
      case(-1)
        select case (is_particle_mdl(particle_in))
        ! ghost
        case (.true.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! colored ghost
          case (.true.)
            particle_type = 'G'
          ! ghost
          case (.false.)
            particle_type = 'x'
          end select
        ! antighost
        case (.false.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! colored antighost
          case (.true.)
            particle_type = 'G~'
          ! antighost
          case (.false.)
            particle_type = 'x~'
          end select
        end select
      ! scalar
      case (1)
        particle_type = 's'
      ! fermion
      case (2)
        select case (is_particle_mdl(particle_in))
        ! fermion
        case (.true.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! quark
          case (.true.)
            particle_type = 'q'
          ! fermion
          case (.false.)
            particle_type = 'f'
          end select
        ! antifermion
        case (.false.)
          select case (get_particle_colour_mdl(particle_in) .ne. 1)
          ! antiquark
          case (.true.)
            particle_type = 'q~'
          ! antifermion
          case (.false.)
            particle_type = 'f~'
          end select
        end select
      ! vector
      case (3)
        select case (get_particle_colour_mdl(particle_in) .ne. 1)
        ! gluon
        case (.true.)
          particle_type = 'g'
        ! vector
        case (.false.)
          particle_type = 'v'
        end select
      case default
        particle_type = '0'
      end select
    end function get_particle_type2_mdl

    function get_axodraw_propagator_mdl(particle_type) result(axoprop)
      character(2), intent(in) :: particle_type
      character(20) :: axoprop
      select case (particle_type)
        case ('g')
          axoprop = '\Gluon'
        case ('v')
          axoprop = '\Photon'
        case ('f', 'f~')
          axoprop = '\ArrowLine'
        case ('q', 'q~')
          axoprop = '\ArrowLine'
        case ('x', 'x~')
          axoprop = '\DashArrowLine'
        case ('G', 'G~')
          axoprop = '\DashLine'
        case ('s')
          axoprop = '\DashLine'
        case default
          write(*,*) "particle_type:", particle_type
          write (*, *) 'Case not implemented in axodraw_propagator'
          stop
      end select
    end function get_axodraw_propagator_mdl
end module class_particles
