!******************************************************************************!
!                                                                              !
!    class_tree.f90                                                            !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2020 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   class_tree.f90                                 |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: J.-N. Lang, S. Rode                    |            !
!        | `---' | UFO date: 06/2020                              |            !
!        |       | UFO version: 0.9                               |            !
!        |       | Generation date: 2020-08-02 21:15             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module class_tree
  use constants_mdl
  use input_mdl, only: error_mdl
  use tcVVV, only: treeVVV
  use tcVVVV, only: treeVVVV
  use tcVVSS, only: treeVVSS
  use tcVSS, only: treeVSS
  use tcFFaV, only: treeFFaV
  use tcUaUV, only: treeUaUV
  use tcVSVS, only: treeVSVS
  use tcVSSV, only: treeVSSV
  use tcSVVS, only: treeSVVS
  use tcSVSV, only: treeSVSV
  use tcSSVV, only: treeSSVV
  use tcSVS, only: treeSVS
  use tcSSV, only: treeSSV
  use tcVFaF, only: treeVFaF
  use tcVFFa, only: treeVFFa
  use tcFaVF, only: treeFaVF
  use tcFaFV, only: treeFaFV
  use tcFVFa, only: treeFVFa
  use tcUaVU, only: treeUaVU
  use tcUUaV, only: treeUUaV
  use tcUVUa, only: treeUVUa
  use tcVUaU, only: treeVUaU
  use tcVUUa, only: treeVUUa
  use tcSS, only: treeSS
  use tcFFa, only: treeFFa
  use tcVV, only: treeVV
  use tcFaF, only: treeFaF
  use tcSSSS, only: treeSSSS
  implicit none

  interface compute_tree_mdl
    module procedure computeTree2, computeTree3, computeTree4, computeTree5,   &
    computeTree6, computeTree7, computeTree8
  end interface 

  contains

  subroutine computeTree2(ty, co, w1, P1, PL1, den, m, wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case(SS0, SS1)
      call treeSS(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case(FFa0, FFa1, FFa2, FFa3)
      call treeFFa(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case(VV0, VV1, VV2)
      call treeVV(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case(FaF0, FaF1, FaF2, FaF3)
      call treeFaF(ty, co, w1, P1, PL1, den, m, wout, last, VB_last)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree2

  subroutine computeTree3(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case(VVV0)
      call treeVVV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(VSS0)
      call treeVSS(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(FFaV0, FFaV1)
      call treeFFaV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(UaUV0)
      call treeUaUV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(SVS0)
      call treeSVS(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(SSV0)
      call treeSSV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,       &
      VB_last)
    case(VFaF0, VFaF1)
      call treeVFaF(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(VFFa0, VFFa1)
      call treeVFFa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(FaVF0, FaVF1)
      call treeFaVF(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(FaFV0, FaFV1)
      call treeFaFV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(FVFa0, FVFa1)
      call treeFVFa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(UaVU0)
      call treeUaVU(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(UUaV0)
      call treeUUaV(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(UVUa0)
      call treeUVUa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(VUaU0)
      call treeVUaU(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case(VUUa0)
      call treeVUUa(ty, co, w1, P1, PL1, w2, P2, PL2, den, m, wout, last,      &
      VB_last)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree3

  subroutine computeTree4(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m&
  , wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case(VVVV0)
      call treeVVVV(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(VVSS0)
      call treeVVSS(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(VSVS0)
      call treeVSVS(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(VSSV0)
      call treeVSSV(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(SVVS0)
      call treeSVVS(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(SVSV0)
      call treeSVSV(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(SSVV0)
      call treeSSVV(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case(SSSS0)
      call treeSSSS(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, den, m, wout&
      , last, VB_last)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree4

  subroutine computeTree5(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, den, m, wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree5

  subroutine computeTree6(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, w5, P5, PL5, den, m, wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in), dimension(0:3) :: P5, PL5
    complex(kind=dp), intent(in), dimension(0:3) :: w5
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree6

  subroutine computeTree7(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, w5, P5, PL5, w6, P6, PL6, den, m, wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in), dimension(0:3) :: P5, PL5
    complex(kind=dp), intent(in), dimension(0:3) :: w5
    complex(kind=dp), intent(in), dimension(0:3) :: P6, PL6
    complex(kind=dp), intent(in), dimension(0:3) :: w6
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree7

  subroutine computeTree8(ty, co, w1, P1, PL1, w2, P2, PL2, w3, P3, PL3, w4, P4&
  , PL4, w5, P5, PL5, w6, P6, PL6, w7, P7, PL7, den, m, wout, last)
    use class_vertices
    integer, intent(in) :: ty
    logical :: VB_last
    logical, intent(in) :: last
    complex(kind=dp), dimension(0:), intent(in) :: co
    complex(kind=dp), intent(in), dimension(0:3) :: P1, PL1
    complex(kind=dp), intent(in), dimension(0:3) :: w1
    complex(kind=dp), intent(in), dimension(0:3) :: P2, PL2
    complex(kind=dp), intent(in), dimension(0:3) :: w2
    complex(kind=dp), intent(in), dimension(0:3) :: P3, PL3
    complex(kind=dp), intent(in), dimension(0:3) :: w3
    complex(kind=dp), intent(in), dimension(0:3) :: P4, PL4
    complex(kind=dp), intent(in), dimension(0:3) :: w4
    complex(kind=dp), intent(in), dimension(0:3) :: P5, PL5
    complex(kind=dp), intent(in), dimension(0:3) :: w5
    complex(kind=dp), intent(in), dimension(0:3) :: P6, PL6
    complex(kind=dp), intent(in), dimension(0:3) :: w6
    complex(kind=dp), intent(in), dimension(0:3) :: P7, PL7
    complex(kind=dp), intent(in), dimension(0:3) :: w7
    complex(kind=dp), intent(in) :: m
    complex(kind=dp), intent(in) :: den
    complex(kind=dp), dimension(0:3), intent(out) :: wout
    character(len=3) :: ty_str
    complex(kind=dp), dimension(0:3) :: pl, p

    wout(:) = 0.0d0
    VB_last = .false.

    select case(ty)
    case default
      write(ty_str,'(i3)') ty
      call error_mdl('Unhandled case: ' // ty_str, where='compute_tree_mdl')
    end select 
    if (last .eqv. .false.) then
      wout(:) = cima * wout(:)/den
    else
      if (VB_last) wout(0) = - wout(0)
    end if 
  end subroutine computeTree8

  subroutine helicity_conservation_mdl(fheT, typeT, fhfilter, j, w, wOut, ml,  &
  last_tl)
    use class_vertices
    integer, dimension(0:), intent(inout) :: fheT
    logical, dimension(:), intent(inout) :: fhfilter
    integer, intent(in), dimension(:) :: w
    integer, intent(in) :: wOut, typeT, j
    logical, intent(in) :: ml, last_tl
    integer, parameter :: pos_hel = 1
    integer, parameter :: neg_hel = -1
    integer, parameter :: unpol = 0
    integer, parameter :: zero_current = 3
    integer, parameter :: not_set = 111

    select case(typeT)
    case(FFaV0)
      if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(FFaV1)
      if (fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. unpol) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(VFaF0)
      if (ml .and. fheT(w(2)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(2)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(VFaF1)
      if (ml .and. fheT(w(2)) .eq. unpol) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(2)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(VFFa0)
      if (ml .and. fheT(w(2)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(2)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(VFFa1)
      if (ml .and. fheT(w(2)) .eq. unpol) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(ml .and. fheT(w(2)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FaVF0)
      if (ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FaVF1)
      if (ml .and. fheT(w(1)) .eq. unpol) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(1)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(FaFV0)
      if (fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(FaFV1)
      if (fheT(w(1)) .eq. unpol .and. fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. pos_hel .and. fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. unpol) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(fheT(w(1)) .eq. neg_hel .and. fheT(w(2)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(FVFa0)
      if (ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FVFa1)
      if (ml .and. fheT(w(1)) .eq. unpol) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(1)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FFa0)
      if (ml .and. fheT(w(1)) .eq. unpol) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(1)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FFa1)
      if (ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FFa2)
      if (ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FFa3)
      if (ml .and. fheT(w(1)) .eq. unpol) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(1)) .eq. pos_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FaF0)
      if (ml .and. fheT(w(1)) .eq. unpol) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(1)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case(FaF1)
      if (ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FaF2)
      if (ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. neg_hel) then
        fheT(wOut) = neg_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else
        fheT(wOut) = unpol
      end if 
    case(FaF3)
      if (ml .and. fheT(w(1)) .eq. unpol) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(ml .and. fheT(w(1)) .eq. pos_hel) then
        fheT(wOut) = pos_hel
        if (last_tl) then
          fheT(wOut) = - fheT(wOut)
        end if 
      else if(fheT(w(1)) .eq. neg_hel) then
        fhfilter(j) = .false.
        if (fheT(wOut) .eq. not_set) fheT(wOut) = zero_current
      else
        fheT(wOut) = unpol
      end if 
    case default
      fheT(wOut) = unpol
    end select 
  end subroutine helicity_conservation_mdl
end module class_tree
