!******************************************************************************!
!                                                                              !
!    lcFaVF.f90                                                                !
!    is part of RECOLA2 (REcursive Computation of One Loop Amplitudes)2        !
!                                                                              !
!    Copyright (C) 2016-2020 Ansgar Denner, Jean-Nicolas Lang and              !
!                            Sandro Uccirati                                   !
!                                                                              !
!    RECOLA2 is licenced under the GNU GPL version 3,                          !
!    see COPYING for details.                                                  !
!                                                                              !
!******************************************************************************!
!                                                           .---.              !
!                                                          /  .  \             !
!                                                         |\_/|   |            !
!                                                         |   |  /|            !
!          .----------------------------------------------------' |            !
!         /  .-.   lcFaVF.f90                                     |            !
!        |  /   \  This code was generated with REPT1L.           |            !
!        | |\_.  | Change this file and you get what you deserve. |            !
!        |\|  | /| Author: J.-N. Lang, S. Rode                    |            !
!        | `---' | UFO date: 06/2020                              |            !
!        |       | UFO version: 0.9                               |            !
!        |       | Generation date: 2020-08-02 21:15             /             !
!        |       |----------------------------------------------'              !
!        \       |                                                             !
!         \     /                                                              !
!          `---'                                                               !
!******************************************************************************!
module lcFaVF
  use class_vertices
  use constants_mdl
  implicit none
  complex (dp) :: x0, x1, x2, x3, x4, x5, x6, x7, x8, x9, x10, x11, x12, x13,  &
                  x14, x15, x16, x17, x18, x19, x20, x21, x22, x23, x24, x25,  &
                  x26, x27, x28, x29
  contains
    subroutine loopFaVF(ty, co, wl, P1, PL1, w2, P2, PL2, m, wp, riMaxIn,      &
    riMaxOut, rankInc)
      use class_vertices
      integer, intent(in) :: riMaxIn, riMaxOut, rankInc
      complex (kind = dp), intent(in), dimension(0:3) :: P1, PL1
      complex (kind = dp), intent(in), dimension(0:3, 0:riMaxIn) :: wl
      complex (kind = dp), intent(in), dimension(0:3) :: P2, PL2
      complex (kind = dp), intent(in), dimension(0:3) :: w2
      complex (dp), intent(out), dimension(0:3,0:riMaxOut) :: wp
      complex (kind = dp), intent(in) :: m
      complex (kind = dp), dimension(0:3) :: pl, p
      integer, intent(in) :: ty
      complex (kind = dp), dimension(0:), intent(in) :: co
      complex (kind = dp), dimension(0:3) :: sp1
      complex (kind = dp), dimension(0:3,0:3) :: sp2
      complex (kind = dp), dimension(0:3,0:3,0:3) :: sp3
      complex (kind = dp), dimension(0:3,0:3,0:3,0:3) :: sp4
      character(len=3) :: ty_str
      integer :: riOut, j
      logical :: isFirstRI
      integer :: mu, nu, ro, ta, te, tu
      select case(ty)
      case(FaVF0)
!       GammaP(2,1,3)*c0
!       GammaM(2,1,3)*c1
        pl(:) = PL1(:) + PL2(:)
        x2 = w2(0) + w2(3)
        x7 = cima*w2(1)
        x8 = w2(0) - w2(3)
        x22 = w2(2) + x7
        x27 = cima*w2(2) + w2(1)
        x29 = co(1)*m
        do j = riMaxIn, 0, -1
          x0 = w2(2)*wl(1,j)
          x1 = w2(1)*wl(1,j)
          x3 = wl(0,j)*x2
          x4 = cima*(x1 + x3)
          x5 = co(1)*(x0 - x4)
          x6 = w2(2)*wl(0,j)
          x9 = wl(1,j)*x8
          x10 = cima*x9
          x11 = co(1)*(wl(0,j)*x7 + x10 + x6)
          x12 = -x11
          x13 = w2(1)*wl(3,j)
          x14 = -w2(0)*wl(2,j) + w2(3)*wl(2,j)
          x15 = x13 + x14
          x16 = w2(2)*wl(3,j)
          x17 = co(0)*(cima*x15 - x16)
          x18 = w2(2)*wl(2,j)
          x19 = wl(3,j)*x2
          x20 = cima*x19
          x21 = co(0)*(wl(2,j)*x7 + x18 - x20)
          x23 = cima*wl(2,j)
          x24 = co(0)*(-wl(3,j)*(-w2(2) + x7) + x23*x8)
          x25 = w2(1)*wl(0,j)
          x26 = x25 + x9
          x28 = w2(1)*wl(2,j)
          wp(0,j) = co(1)*(cima*pl(2)*x26 + pl(1)*w2(2)*wl(1,j) - pl(1)*x4 + pl&
                    (2)*w2(2)*wl(0,j)) + m*x17
          wp(1,j) = -co(1)*(cima*pl(0)*(w2(0)*wl(1,j) - w2(3)*wl(1,j) + x25) - &
                    cima*pl(3)*(x1 + x3) + pl(0)*w2(2)*wl(0,j) + pl(3)*w2(2)*wl&
                    (1,j)) + m*x21
          wp(2,j) = co(0)*(cima*pl(0)*x15 - pl(0)*w2(2)*wl(3,j) + pl(2)*(-cima*&
                    x19 + cima*x28 + w2(2)*wl(2,j))) + x29*(x0 - x4)
          wp(3,j) = co(0)*(cima*pl(3)*x13 + pl(1)*(w2(1)*x23 + x18 - x20) - pl(&
                    3)*(x16 + x23*x8)) - x29*(wl(0,j)*x22 + x10)
          if(rankInc .lt. 1) cycle
          sp2(0,0) = x5
          sp2(0,1) = x12
          sp2(0,2) = x17
          sp2(0,3) = x21
          sp2(1,0) = x12
          sp2(1,1) = x5
          sp2(1,2) = co(0)*(-wl(2,j)*x22 + x20)
          sp2(1,3) = x24
          sp2(2,0) = co(1)*(-cima*x6 + x26)
          sp2(2,1) = -co(1)*(wl(1,j)*x27 + x3)
          sp2(2,2) = -co(0)*(cima*x18 + x19 - x28)
          sp2(2,3) = -co(0)*(wl(3,j)*x27 + x14)
          sp2(3,0) = x5
          sp2(3,1) = x11
          sp2(3,2) = x24
          sp2(3,3) = x21
          do mu = 0, 3
            riOut = incRI(mu, j)
            wp(:,riOut) = wp(:,riOut) + sp2(mu,:)
          end do 
        end do 
      case(FaVF1)
!       GammaM(2,1,3)*c0
        pl(:) = PL1(:) + PL2(:)
        do j = riMaxIn, 0, -1
          x0 = w2(2)*wl(1,j)
          x1 = w2(1)*wl(1,j)
          x2 = wl(0,j)*(w2(0) + w2(3))
          x3 = cima*(x1 + x2)
          x4 = co(0)*(x0 - x3)
          x5 = w2(2)*wl(0,j)
          x6 = w2(1)*wl(0,j)
          x7 = wl(1,j)*(w2(0) - w2(3))
          x8 = co(0)*(cima*x6 + cima*x7 + x5)
          x9 = -x8
          x10 = x6 + x7
          wp(0,j) = co(0)*(cima*pl(2)*x10 + pl(1)*w2(2)*wl(1,j) - pl(1)*x3 + pl&
                    (2)*w2(2)*wl(0,j))
          wp(1,j) = co(0)*(-cima*pl(0)*(w2(0)*wl(1,j) - w2(3)*wl(1,j) + x6) +  &
                    cima*pl(3)*(x1 + x2) - w2(2)*(pl(0)*wl(0,j) + pl(3)*wl(1,j &
                    )))
          wp(2,j) = co(0)*m*(x0 - x3)
          wp(3,j) = -m*x8
          if(rankInc .lt. 1) cycle
          sp2(0,0) = x4
          sp2(0,1) = x9
          sp2(0,2) = 0
          sp2(0,3) = 0
          sp2(1,0) = x9
          sp2(1,1) = x4
          sp2(1,2) = 0
          sp2(1,3) = 0
          sp2(2,0) = co(0)*(-cima*x5 + x10)
          sp2(2,1) = -co(0)*(cima*w2(2)*wl(1,j) + w2(1)*wl(1,j) + x2)
          sp2(2,2) = 0
          sp2(2,3) = 0
          sp2(3,0) = x4
          sp2(3,1) = x8
          sp2(3,2) = 0
          sp2(3,3) = 0
          do mu = 0, 3
            riOut = incRI(mu, j)
            wp(:,riOut) = wp(:,riOut) + sp2(mu,:)
          end do 
        end do 
      end select 
    end subroutine loopFaVF
end module lcFaVF
